package ac.ed.lurg.country;

import java.util.HashMap;
import java.util.Map;

import ac.ed.lurg.ModelConfig;
import ac.ed.lurg.country.gams.GamsCountryInput;
import ac.ed.lurg.country.gams.GamsRasterInput;
import ac.ed.lurg.country.gams.GamsRasterOptimiser;
import ac.ed.lurg.country.gams.GamsRasterOutput;
import ac.ed.lurg.demand.DemandManager;
import ac.ed.lurg.landuse.AreasItem;
import ac.ed.lurg.landuse.CropUsageData;
import ac.ed.lurg.landuse.IrrigationCostItem;
import ac.ed.lurg.landuse.LandCoverItem;
import ac.ed.lurg.types.CommodityType;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.YieldRaster;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class CountryAgent {
	
	private DemandManager demandManager;
	private Country country;
	
	private Map<Integer, GamsRasterOutput> resultsTimeseries = new HashMap<Integer, GamsRasterOutput>();

	private int currentTimestep;
	private YieldRaster countryYieldSurfaces;
	private RasterSet<IrrigationCostItem> irrigationCostRaster;

	public CountryAgent(DemandManager demandManager, Country country, RasterSet<LandCoverItem> initialLC,
			RasterSet<IrrigationCostItem> irrigationCostRaster, Map<CropType, CropUsageData> cropUsageData) {
		
		this.demandManager = demandManager;
		this.country = country;
		RasterSet<AreasItem> cropAreaRaster = convertInitialLC(initialLC);
		this.irrigationCostRaster = irrigationCostRaster;
		
		GamsRasterOutput initialData = new GamsRasterOutput(cropAreaRaster, cropUsageData);
		
		resultsTimeseries.put(0, initialData);
	}

	private RasterSet<AreasItem> convertInitialLC(RasterSet<LandCoverItem> initialLC) {
		RasterSet<AreasItem> cropAreaRaster = new RasterSet<AreasItem>(initialLC.getHeaderDetails());
		
		for (Map.Entry<RasterKey, LandCoverItem> entry : initialLC.entrySet()) {
			AreasItem areasItem = new AreasItem();
			RasterKey key = entry.getKey();
			areasItem.setLandCoverAreas(entry.getValue());
			areasItem.setCropFraction(CropType.WHEAT, 0.5); // random start, better if we could get data, but free substitution between crops so not critical
			areasItem.setCropFraction(CropType.MAIZE, 0.5);
			cropAreaRaster.put(key, areasItem);
		}
		
		return cropAreaRaster;
	}
	
	public Country getCountry() {
		return country;
	}

	public GamsRasterOutput determineProduction(int timestep, YieldRaster countryYieldSurfaces, Map<CropType, Double> worldInputEnergy) {
		currentTimestep = timestep;
		this.countryYieldSurfaces = countryYieldSurfaces;
		
		// get projected demand
		int year = ModelConfig.BASE_YEAR + currentTimestep;
		Map<CommodityType, Double> projectedDemand = demandManager.getDemand(country, year);
		
		if (projectedDemand.size() == 0) {
			LogWriter.printlnError("No demand for country " + country + " so skipping it");
		}
		else if (countryYieldSurfaces.size() == 0 ) {
			LogWriter.printlnError("No yield values for " + country + " so skipping it");
		}
		else {
			// optimize areas and intensity 
			GamsRasterInput input = getGamsRasterInput(projectedDemand, worldInputEnergy);
			GamsRasterOptimiser opti = new GamsRasterOptimiser(input);
			LogWriter.println("Running " + country.getCountryName() + ", currentTimestep " + currentTimestep);
			
			GamsRasterOutput result = opti.run();
			resultsTimeseries.put(timestep, result);
			return result;
		}
		
		return null;
	}

	private GamsRasterInput getGamsRasterInput(Map<CommodityType, Double> projectedDemand, Map<CropType, Double> worldInputEnergy) {

		GamsRasterOutput prevOutput;
		Map<CropType, Double> cropAdjs;
		boolean calibrate;

		if (ModelConfig.START_TIMESTEP == currentTimestep) {  // initialisation time-step
			calibrate = true;
			prevOutput = resultsTimeseries.get(currentTimestep);
			cropAdjs = null;
		}
		else { // normal (not the initial) time-step
			calibrate = false;
			prevOutput = resultsTimeseries.get(currentTimestep - 1);
			cropAdjs = prevOutput.getCropAdjustments();
		}
		
		Map<CropType, Double> baseNetImport = new HashMap<CropType, Double>();
		Map<CropType, Double> maxOfProdOrSupply = new HashMap<CropType, Double>();

		for (Map.Entry<CropType, CropUsageData> entry : prevOutput.getCropUsageData().entrySet()) {
			CropUsageData cropUsage = entry.getValue();
			double netImports = cropUsage.getNetImports();
			baseNetImport.put(entry.getKey(), netImports);
			maxOfProdOrSupply.put(entry.getKey(), cropUsage.getProduction() + Math.max(netImports, 0));
		}
		
		GamsCountryInput countryLevelInputs = GamsCountryInput.createInput(country, projectedDemand, worldInputEnergy, baseNetImport, maxOfProdOrSupply, cropAdjs, calibrate);	
		GamsRasterInput input = new GamsRasterInput(countryYieldSurfaces, prevOutput.getCropAreaRaster(), irrigationCostRaster, countryLevelInputs);

		return input;
	}
}