package ac.ed.lurg.country.gams;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;

import ac.ed.lurg.landuse.AreasItem;
import ac.ed.lurg.landuse.IntensitiesItem;
import ac.ed.lurg.landuse.IrrigationCostItem;
import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.LandCoverType;
import ac.ed.lurg.types.YieldType;
import ac.ed.lurg.utils.LazyHashMap;
import ac.ed.lurg.utils.LogWriter;
import ac.ed.lurg.yield.AveragingYieldResponsesItem;
import ac.ed.lurg.yield.YieldRaster;
import ac.ed.lurg.yield.YieldResponsesItem;
import ac.sac.raster.IntegerRasterItem;
import ac.sac.raster.RasterKey;
import ac.sac.raster.RasterSet;

public class GamsRasterOptimiser {
	public static final boolean DEBUG = false;

	private GamsRasterInput rasterInputData;
	private LazyHashMap<Integer, Set<RasterKey>> mapping;

	public GamsRasterOptimiser(GamsRasterInput rasterInputData) {
		this.rasterInputData = rasterInputData;
	}

	public GamsRasterOutput run() {
		// workout similar areas		
		GamsLocationInput gamsInput = convertFromRaster(rasterInputData);

		// run optimizer
		GamsLocationOptimiser opti = new GamsLocationOptimiser(gamsInput);		
		GamsLocationOutput gamsOutput = opti.run();

		// map results back to raster
		return convertToRaster(gamsInput, gamsOutput);
	}

	private GamsRasterOutput convertToRaster(GamsLocationInput gamsInput, GamsLocationOutput gamsOutput) {		
		RasterSet<AreasItem> newAreaRaster = allocAreas(gamsInput.getPreviousAreas(), gamsOutput);
		RasterSet<IntensitiesItem> newIntensityRaster = allocIntensities(gamsOutput);

		RasterSet<IntegerRasterItem> locationIdRaster = new RasterSet<IntegerRasterItem>(rasterInputData.getPreviousAreas().getHeaderDetails());
		for (Entry<Integer, Set<RasterKey>> entry : mapping.entrySet()) {
			Integer locId = entry.getKey();
			for (RasterKey key : entry.getValue())
				locationIdRaster.put(key, new IntegerRasterItem(locId));
		}
		
		return new GamsRasterOutput(gamsOutput.getStatus(), newIntensityRaster, newAreaRaster, gamsOutput.getCommoditiesData(), gamsOutput.getCropAdjustments(), locationIdRaster);
	}

	private RasterSet<AreasItem> createWithSameLandCovers(RasterSet<AreasItem> toCopy) {
		RasterSet<AreasItem> theCopy = new RasterSet<AreasItem>(toCopy.getHeaderDetails());
		
		for (Entry<RasterKey, AreasItem> entry : toCopy.entrySet()) {
			AreasItem newAreasItem = new AreasItem();
			newAreasItem.setLandCoverAreas(entry.getValue());
			theCopy.put(entry.getKey(), newAreasItem);
		}
		
		return theCopy;
	}
	
	@SuppressWarnings("unused")
	private void checkedTotalAreas(RasterSet<AreasItem> areaRaster, String comment) {
		for (LandCoverType l : LandCoverType.values()) {
			double total = 0;
			for (AreasItem a : areaRaster.values()) {
				total += a.getLandCoverArea(l);
			}
			
			LogWriter.printlnError("Total Area " + comment + ": " + l.getName() + ": " + total);
		}
	}
	
	private RasterSet<AreasItem> allocAreas(Map<Integer, ? extends AreasItem> prevAreasAgg, GamsLocationOutput gamsOutput) {
		RasterSet<AreasItem> newAreaRaster = createWithSameLandCovers(rasterInputData.getPreviousAreas());

		//checkedTotalAreas(rasterInputData.getPreviousAreas(), "old");
		//checkedTotalAreas(newAreaRaster, "new");

		for (Map.Entry<Integer, AreasItem> entry : gamsOutput.getCropAreas().entrySet()) {
			Integer locId = entry.getKey();
			AreasItem newAreaAggItem = entry.getValue();
			AreasItem prevAreaAggItem = prevAreasAgg.get(locId);
			Set<RasterKey> keys = mapping.get(locId);

			//checkedTotalAreas(newAreaRaster.popSubsetForKeys(new RasterSet<AreasItem>(), keys), locId + " setset");

			
			// if (DEBUG) LogWriter.println("Processing location id " + locId);
			
			double pastureChange = newAreaAggItem.getLandCoverArea(LandCoverType.PASTURE) - prevAreaAggItem.getLandCoverArea(LandCoverType.PASTURE);
			double croplandChange = newAreaAggItem.getLandCoverArea(LandCoverType.CROPLAND) - prevAreaAggItem.getLandCoverArea(LandCoverType.CROPLAND);
			double prevForestToNaturalFraction = prevAreaAggItem.getForestToNaturalFraction();

			double pastureFromCrop = 0;
			double pastureFromNatural = 0;
			double cropFromNatural = 0;

			if (pastureChange > 0 && croplandChange < 0) {
				pastureFromCrop = Math.min(pastureChange, -croplandChange);
				pastureFromNatural = (pastureChange > -croplandChange) ? pastureChange+croplandChange : 0;
				cropFromNatural = croplandChange + pastureFromCrop;
			}
			else if (pastureChange < 0 && croplandChange > 0) {
				pastureFromCrop = -Math.min(-pastureChange, croplandChange);
				cropFromNatural = (croplandChange > -pastureChange) ? -(pastureChange+croplandChange) : 0;
				pastureFromNatural = pastureChange - pastureFromCrop;
			}
			else {
				pastureFromNatural = pastureChange;
				cropFromNatural = croplandChange;
			}
			
			double shortfall = 0;
			shortfall += allocAllLandCrops(newAreaRaster, keys, LandCoverType.PASTURE, LandCoverType.FOREST, pastureFromNatural * prevForestToNaturalFraction);
			shortfall += allocAllLandCrops(newAreaRaster, keys, LandCoverType.PASTURE, LandCoverType.OTHER_NATURAL, pastureFromNatural * (1-prevForestToNaturalFraction));
			shortfall += allocAllLandCrops(newAreaRaster, keys, LandCoverType.PASTURE, LandCoverType.CROPLAND, pastureFromCrop);
			shortfall += allocAllLandCrops(newAreaRaster, keys, LandCoverType.CROPLAND, LandCoverType.FOREST, cropFromNatural * prevForestToNaturalFraction);
			shortfall += allocAllLandCrops(newAreaRaster, keys, LandCoverType.CROPLAND, LandCoverType.OTHER_NATURAL, cropFromNatural * (1-prevForestToNaturalFraction));

			if (shortfall > 0.00001)
				LogWriter.printlnError("This should never happen, due to GAMS constraint. Not able to incorporate all changes, as not enough forest or natural areas left: " + shortfall);

			for (RasterKey key : keys) {
				AreasItem newAreasRasterItem = newAreaRaster.get(key);
				for (CropType crop : CropType.values()) {
					newAreasRasterItem.setCropFraction(crop, newAreaAggItem.getCropFraction(crop));
				}
			}
		}

		return newAreaRaster;
	}
	
	private double allocAllLandCrops(RasterSet<AreasItem> newAreaRaster, Set<RasterKey> keys, LandCoverType toLC, LandCoverType fromLC, double change) {
		LandCoverType toType = toLC;
		LandCoverType fromType = fromLC;
		
		// reverse direction if negative
		if (change < 0) {
			change = -change;
			toType = fromLC;
			fromType = toLC;
		}
		
		double totalShortfall = 0;
		Set<RasterKey> keysWithSpace = new HashSet<RasterKey>();
		double avgChange = change/keys.size();

		for (RasterKey key : keys) {
			//if (DEBUG) LogWriter.println("  Processing raster key " + key);
			AreasItem newAreasRasterItem = newAreaRaster.get(key);
			
			double shortfall = newAreasRasterItem.moveAreas(toType, fromType, avgChange);
			if (shortfall == 0)
				keysWithSpace.add(key);
			else
				totalShortfall += shortfall;
		}

		if (totalShortfall > 0 & keysWithSpace.size() > 0) {  // more to allocate and some free areas to allocate into
			if (DEBUG) LogWriter.println("A total land cover of " + totalShortfall + " is unallocated, so trying again with areas with space");
			totalShortfall = allocAllLandCrops(newAreaRaster, keysWithSpace, toType, fromType, totalShortfall);
		}

		return totalShortfall;
	}

	private RasterSet<IntensitiesItem> allocIntensities(GamsLocationOutput gamsOutput) {

		RasterSet<IntensitiesItem> newIntensityRaster = new RasterSet<IntensitiesItem>(rasterInputData.getPreviousAreas().getHeaderDetails());

		for (Map.Entry<Integer, AreasItem> entry : gamsOutput.getCropAreas().entrySet()) {
			Integer locId = entry.getKey();
			IntensitiesItem newIntensitiesItem = gamsOutput.getIntensities().get(locId);

			for (RasterKey key : mapping.get(locId)) {
				newIntensityRaster.put(key, newIntensitiesItem); // intensities constant over single aggregated land category
			}
		}
		return newIntensityRaster;
	}

	@SuppressWarnings("serial")
	private GamsLocationInput convertFromRaster(GamsRasterInput rasterInputData) {
		// as a first attempt only going to look at pasture and cereal yields, assuming other yields will be correlated to one or the other.
		YieldRaster yieldRaster = rasterInputData.getYields();
		RasterSet<AreasItem> cropAreaRaster = rasterInputData.getPreviousAreas();
		
	/*	checkedTotalAreas(cropAreaRaster, "before");
		new RasterOutputer<AreasItem>(cropAreaRaster, "BeforeLandCover") {
			@Override
			public Double getValue(RasterKey location) {
				AreasItem area = results.get(location);
				if (area == null)
					return null;

				int count = 0;
				double maxLC = 0;
				int maxId = 0;
				for (LandCoverType l : LandCoverType.values()) {
					count++;
					if (maxLC < area.getLandCoverArea(l)) {
						maxLC = area.getLandCoverArea(l);
						maxId = count;
					}
				}
				return (double)maxId;
			}

			@Override
			public int convertToPixelValue(double value) {
				return (int) (1 + value * 2);
			}
		}.writeOutput(true); */


		RasterSet<IrrigationCostItem> irrigCostRaster = rasterInputData.getIrrigationCost();

		{
				//	YieldResponsesItem yresp = yieldRaster.get(new RasterKey(40,38));//getFromCoordinates(-118.0, -33.0);
				//	LogWriter.printlnError("Test key2: " + yresp.getYieldMax(CropType.WHEAT)  + ", " + yresp.getYieldFertOnly(CropType.WHEAT) + ", " + yresp.getYieldIrrigOnly(CropType.WHEAT));
		}

		// look for inconsistencies
		for (Entry<RasterKey, YieldResponsesItem> entry : yieldRaster.entrySet()) {

			YieldResponsesItem yresp = entry.getValue();
			RasterKey key = entry.getKey();
			CropType crop = CropType.WHEAT;

			if (yresp != null && (yresp.getYieldMax(crop) < yresp.getYieldFertOnly(crop) || yresp.getYieldMax(crop) < yresp.getYieldIrrigOnly(crop))) {
				LogWriter.printlnError(key.getCol() + ", "  + key.getRow() + ": " + yieldRaster.getXCoordin(key) + ", " + yieldRaster.getYCoordin(key) + ": " + yresp.getYieldMax(crop)  + ", " + yresp.getYieldFertOnly(crop) + ", " + yresp.getYieldIrrigOnly(crop));
			}
			//	else if (yresp != null)
			//		LogWriter.println(key.getCol() + ", "  + key.getRow() + ": " + yieldRaster.getXCoordin(key) + ", " + yieldRaster.getYCoordin(key) + ": " + yresp.getYieldMax(CropType.CEREALS)  + ", " + yresp.getYieldFertOnly(CropType.CEREALS) + ", " + yresp.getYieldIrrigOnly(CropType.CEREALS));

		}

		int numCerealCats = 5;
		int numPastureCats = 1;

		int thisShouldLookAtCropsOtherThanJustWheat; // need to consider other crops, and perhaps other yieldTypes as well
		List<Double> wheatlDivisions = getDivisions(yieldRaster, CropType.WHEAT, numCerealCats); 
		List<Double> pastureDivisions = getDivisions(yieldRaster, CropType.PASTURE, numPastureCats);

		if (DEBUG) LogWriter.println("Making " + numCerealCats * numPastureCats + " categories");

		LazyHashMap<Integer, AveragingYieldResponsesItem> aggregatedYields = new LazyHashMap<Integer, AveragingYieldResponsesItem>() { 
			protected AveragingYieldResponsesItem createValue() { return new AveragingYieldResponsesItem(); }
		};
		LazyHashMap<Integer, AreasItem> aggregatedAreas = new LazyHashMap<Integer, AreasItem>() { 
			protected AreasItem createValue() { return new AreasItem(); }
		};
		LazyHashMap<Integer, IrrigationCostItem> aggregatedIrrigCosts = new LazyHashMap<Integer, IrrigationCostItem>() { 
			protected IrrigationCostItem createValue() { return new IrrigationCostItem(); }
		};
		mapping = new LazyHashMap<Integer, Set<RasterKey>>() { 
			protected Set<RasterKey> createValue() { return new HashSet<RasterKey>(); }
		};

		int countFound = 0, countMissing = 0;

		
		
		for (Entry<RasterKey, YieldResponsesItem> entry : yieldRaster.entrySet()) {
			YieldResponsesItem yresp = entry.getValue();
			RasterKey key = entry.getKey();

			if (yresp == null) {
				countMissing++;
				LogWriter.printlnError("No YieldResponsesItem for " + key + ", x=" + yieldRaster.getXCoordin(key) + ", y=" + yieldRaster.getYCoordin(key));
			}
			else {
				countFound++;
				//LogWriter.println("YieldResponsesItem found for " + key + ", x=" + yieldRaster.getXCoordin(key) + ", y=" + yieldRaster.getYCoordin(key));

				AreasItem cropAreas  = cropAreaRaster.get(key);
				if (cropAreas == null) {
					LogWriter.printlnError("GamsRasterOptimiser: Can't find cropAreas for " + key);
					continue;
				}
				
				IrrigationCostItem irrigCost = irrigCostRaster.get(key);

				int cerealCat = findCategory(wheatlDivisions, yresp.getYieldMax(CropType.WHEAT));
				int pastureCat = findCategory(pastureDivisions, yresp.getYieldMax(CropType.PASTURE));
				Integer id = cerealCat + pastureCat * numCerealCats;

				AveragingYieldResponsesItem avgYResp = aggregatedYields.lazyGet(id);
				AreasItem aggAreas = aggregatedAreas.lazyGet(id);
				IrrigationCostItem avgIrigCost = aggregatedIrrigCosts.lazyGet(id);
				mapping.lazyGet(id).add(key); 

				// Irrigation cost
				double suitableAreaThisTime  = cropAreas.getSuitableLand();
				if (irrigCost!= null && suitableAreaThisTime > 0) {
					double areaSoFar = aggAreas.getSuitableLand();
					avgIrigCost.setIrrigCost((avgIrigCost.getIrrigCost()*areaSoFar + suitableAreaThisTime*irrigCost.getIrrigCost()) / (areaSoFar + suitableAreaThisTime));
					//LogWriter.println(id + ", " + key + ", " + avgIrigCost.getIrrigCost() + ", from " + areaSoFar + ", " + suitableAreaThisTime + ", " + irrigCost.getIrrigCost());
				}

				// Crops yields and area fractions
				for (CropType crop : CropType.getNonMeatTypes()) {
					for (YieldType yieldType : YieldType.values()) {
						avgYResp.setYield(yieldType, crop, yresp.getYield(yieldType, crop));
					}

					double areaSoFar = aggAreas.getCropArea(crop);
					double areaThisTime = cropAreas.getCropArea(crop);
					double updateCropland = (cropAreas.getLandCoverArea(LandCoverType.CROPLAND) + aggAreas.getLandCoverArea(LandCoverType.CROPLAND));
					
					if (updateCropland > 0)
						aggAreas.setCropFraction(crop, (areaSoFar + areaThisTime) / updateCropland);
				}
				
				// Land covers ares
				for (LandCoverType landType : LandCoverType.values()) {
					double areaThisTime = cropAreas.getLandCoverArea(landType);
					double areaSoFar = aggAreas.getLandCoverArea(landType);
					aggAreas.setLandCoverArea(landType, areaSoFar + areaThisTime);
				}
			}
		}

		LogWriter.println("YieldResponsesItem: " + rasterInputData.getCountryInput().getCountry() + ", countFound=" + countFound + ", countMissing=" + countMissing);

		for (Map.Entry<Integer, Set<RasterKey>> e : mapping.entrySet()) {
			LogWriter.println(e.getKey() + " category has " + e.getValue().size() + " raster areas");
			if (DEBUG) for (RasterKey key : e.getValue()) {
				LogWriter.println(String.format("\t%s: yields pasture=%.1f, cereal=%.1f", key, yieldRaster.get(key).getYieldMax(CropType.PASTURE), yieldRaster.get(key).getYieldMax(CropType.WHEAT)));
			}
		}

		return new GamsLocationInput(aggregatedYields, aggregatedAreas, aggregatedIrrigCosts, rasterInputData.getCountryInput());
	}

	private int findCategory(List<Double> divisions, double yield) {
		int category;
		int numDivisions = divisions.size();

		for (category = 0; category<numDivisions; category++) {
			if (yield < divisions.get(category)) {
				break;
			}
		}
		return category;
	}

	private List<Double> getDivisions(YieldRaster yieldRaster, CropType crop, int numCategories) {
		List<Double> yieldValues = new ArrayList<Double>();

		for (YieldResponsesItem yresp : yieldRaster.values()) {
			if (yresp == null) {
				if (DEBUG) LogWriter.println("GamsRasterOptimiser: Can't get value for crop " + crop);
			}
			else {
				double d = yresp.getYieldMax(crop);
				//LogWriter.println("GamsRasterOptimiser: Got value for crop " + crop + " of " + d);

				if (Double.isNaN(d) || d == 0.0) {
					if (DEBUG) LogWriter.println("GamsRasterOptimiser: Got NaN or zero value for crop " + crop);
				}
				else {
					yieldValues.add(d);
				}
			}
		}

		if (yieldValues.size() == 0) {
			throw new RuntimeException("No yield values for country, crop = " + crop);
		}

		Collections.sort(yieldValues);

		List<Double> divisions = new ArrayList<Double>();

		for (int i=1; i < numCategories; i++) {
			double d = yieldValues.get(yieldValues.size()/numCategories*i);
			divisions.add(d);
		}	

		return divisions;
	}
}
