package ac.ed.lurg.yield;

import java.io.BufferedReader;
import java.io.File;
import java.io.FileReader;

import ac.ed.lurg.types.CropType;
import ac.ed.lurg.types.FertiliserRate;
import ac.ed.lurg.types.IrrigationRate;
import ac.ed.lurg.types.YieldType;
import ac.ed.lurg.utils.LogWriter;
import ac.sac.raster.RasterHeaderDetails;

public class LPJYieldResponseMapReader {	

	private YieldRaster dataset;
	private String rootDir;

	public LPJYieldResponseMapReader(String rootDir, RasterHeaderDetails rasterProj) {
		this.rootDir = rootDir;
		dataset = new YieldRaster(rasterProj);

	}
	
	private String[] parseLine(String line) {
		return line.split(" +");
	}

	public YieldRaster getRasterDataFromFile(FertiliserRate fr) {
		long startTime = System.currentTimeMillis();
		YieldType noIrrigYieldType = YieldType.getYieldType(fr, IrrigationRate.NO_IRRIG);  // just do this look up once per file
		YieldType maxIrrigYieldType = YieldType.getYieldType(fr, IrrigationRate.MAX_IRRIG);

		int col=0;
		String filename = rootDir + File.separator + fr.getFileName();
				
		try {
			BufferedReader in = new BufferedReader(new FileReader(filename));
			String line;
			in.readLine(); // read header
			YieldRecord record = null;

			while ((line=in.readLine()) != null) {
				String[] tokens = parseLine(line.trim());
				record = new YieldRecord();

				for (String token : tokens) {
					double d = Double.parseDouble(token);

					switch (col) {
					case 0:
						record.x = d;
						break;
					case 1:
						record.y = d;
						break;
					case 2:
						//YieldRecord values in kg DM / m2, we want t/ha, so 10 times larger
						record.c3Pasture = d * 10;
						break;
					case 3:
						record.c4Pasture = d * 10;
						break;
					case 4:
						record.teWW = d * 10;
						break;
					case 5:
						record.teSW = d * 10;
						break;
					case 6:
						record.teCo = d * 10;
						break;
					case 7:
						record.trRi = d * 10;
						break;
					case 8:
						record.teWWirr = d * 10;
						break;
					case 9:
						record.teSWirr = d * 10;
						break;
					case 10:
						record.teCoirr = d * 10;
						break;
					case 11:
						record.trRiirr = d * 10;
						break;
					}

					col++;
				}
				YieldResponsesItem data = dataset.getFromCoordinates(record.x, record.y);
				setData(noIrrigYieldType, maxIrrigYieldType, data, record);
				col=0;
			}
			in.close();
		}
		catch (Exception e) {
			LogWriter.printlnError("Problem reading regional yield data: col:" + col);
			e.printStackTrace();
		}

		LogWriter.println("Reading " + filename + ", took " + (System.currentTimeMillis() - startTime) + " ms");

		return dataset;
	}
	
	
	public YieldRaster getYieldRaster() {
		return dataset;
	}
	
	public void setData(YieldType noIrrigYieldType, YieldType maxIrrigYieldType, YieldResponsesItem item, YieldRecord record) {
		
		boolean isSpringWheat = (record.teSW > record.teWW);
		item.setIsSpringWheat(isSpringWheat);
		
		item.setYield(noIrrigYieldType, CropType.WHEAT, isSpringWheat ? record.teSW : record.teWW);
		item.setYield(maxIrrigYieldType, CropType.WHEAT, isSpringWheat ? record.teSWirr : record.teWWirr);
		item.setYield(noIrrigYieldType, CropType.MAIZE, record.teCo);
		item.setYield(maxIrrigYieldType, CropType.MAIZE, record.teCoirr);
		item.setYield(noIrrigYieldType, CropType.RICE, record.trRi);
		item.setYield(maxIrrigYieldType, CropType.RICE, record.trRiirr);
		
		item.setYield(noIrrigYieldType, CropType.TROPICAL_CEREALS, record.teCo);
		item.setYield(maxIrrigYieldType, CropType.TROPICAL_CEREALS, record.teCoirr);
		item.setYield(noIrrigYieldType, CropType.OILCROPS, record.teWW);
		item.setYield(maxIrrigYieldType, CropType.OILCROPS, record.teWWirr);
		item.setYield(noIrrigYieldType, CropType.SOYBEAN, record.teSW);
		item.setYield(maxIrrigYieldType, CropType.SOYBEAN, record.teSWirr);
		item.setYield(noIrrigYieldType, CropType.PULSES, record.teSW);
		item.setYield(maxIrrigYieldType, CropType.PULSES, record.teSWirr);
		item.setYield(noIrrigYieldType, CropType.STARCHY_ROOTS, record.teSW);
		item.setYield(maxIrrigYieldType, CropType.STARCHY_ROOTS, record.teSWirr);
		
		item.setYield(noIrrigYieldType, CropType.PASTURE, Math.max(record.c3Pasture + record.c4Pasture, 0));
		item.setYield(maxIrrigYieldType, CropType.PASTURE,  Math.max(record.c3Pasture + record.c4Pasture, 0)*1.2);
	}
	
	class YieldRecord {
		double x;
		double y;
		double c3Pasture;
		double c4Pasture;
		double teSW;
		double teSWirr;
		double teWW;
		double teWWirr;
		double teCo;
		double teCoirr;
		double trRi;
		double trRiirr;
	}
}